/////////////////////////////////////////////////////////////////////////////
//
// gsvga - Emulate a 320x240x8bpp color-indexed framebuffer
//
/////////////////////////////////////////////////////////////////////////////

#include <eekernel.h>
#include <eeregs.h>
#include <eestruct.h>
#include <libgraph.h>
#include <libdma.h>
#include <sifdev.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <libdev.h>

#include "ps2port.h"
#include "filecache.h"

/////////////////////////////////////////////////////////////////////////////

static sceGsDBuff db;

typedef union {
  u_long128 ul128;
  u_long    ul64[2];
  u_int     ui32[4];
} Q_WORDDATA;

static Q_WORDDATA packet_buf[16];
static sceDmaChan *d2;

#define INDEXIMG_ADDRESS 6720
#define CLUT_ADDRESS     8000

/////////////////////////////////////////////////////////////////////////////
//
// Initialize a packet for drawing the screen texture
//
static void init_packet(Q_WORDDATA *pbuf) {
  pbuf[0].ul128 = (u_long128)0;
  pbuf[0].ui32[0] = 0x1000000d;           // DMA cnt

  pbuf[1].ul64[0] = SCE_GIF_SET_TAG(5, 0, 0, 0, 0, 1);
  pbuf[1].ul64[1] = 0xe;

  pbuf[2].ul64[0] = 0;
  pbuf[2].ul64[1] = SCE_GS_TEXFLUSH;

  pbuf[4].ul64[1]= SCE_GS_TEX0_1;
  pbuf[4].ul64[0]=SCE_GS_SET_TEX0(INDEXIMG_ADDRESS, 6, SCE_GS_PSMT8, 9, 8, 1, 1, CLUT_ADDRESS, SCE_GS_PSMCT32, 0, 0, 1);

  pbuf[5].ul64[0] = SCE_GS_SET_TEX1(0, 0, SCE_GS_NEAREST, SCE_GS_NEAREST, 0, 0, 0);
  pbuf[5].ul64[1] = SCE_GS_TEX1_1;

  pbuf[6].ul64[0] = SCE_GS_SET_CLAMP(1, 1, 0, 0, 0, 0);
  pbuf[6].ul64[1] = SCE_GS_CLAMP_1;

  pbuf[7].ul64[0] = SCE_GIF_SET_TAG(1, 1, 1, SCE_GS_PRIM_SPRITE | SCE_GS_PRIM_TME, 0, 6);
  pbuf[7].ul64[1] = 0x412412;

  pbuf[8].ui32[0] = 0;                    // S
  pbuf[8].ui32[1] = 0;                    // T
  pbuf[8].ui32[2] = 0x3f800000;           // Q(1.0f)
  pbuf[8].ui32[3] = 0;

  pbuf[9].ui32[0] = 0xff;                 // R
  pbuf[9].ui32[1] = 0x40;                 // G
  pbuf[9].ui32[2] = 0x20;                 // B
  pbuf[9].ui32[3] = 0x80;                 // A

  pbuf[10].ui32[0] = (2048 - 160) << 4;   // X
  pbuf[10].ui32[1] = (2048 - 112) << 4;   // Y
  pbuf[10].ui32[2] = 1 << 4;              // Z
  pbuf[10].ui32[3] = 0;                   // F

  pbuf[11].ui32[0] = 0x3f800000;          // S
  pbuf[11].ui32[1] = 0x3f800000;          // T
  pbuf[11].ui32[2] = 0x3f800000;          // Q(1.0f)
  pbuf[11].ui32[3] = 0;

  pbuf[12].ui32[0] = 0xff;                // R
  pbuf[12].ui32[1] = 0x40;                // G
  pbuf[12].ui32[2] = 0x20;                // B
  pbuf[12].ui32[3] = 0x80;                // A

  pbuf[13].ui32[0] = (2048 + 352) << 4;   // X
  pbuf[13].ui32[1] = (2048 + 144) << 4;   // Y
  pbuf[13].ui32[2] = 1 << 4;              // Z
  pbuf[13].ui32[3] = 0;                   // F

  pbuf[14].ul128 = (u_long128)0;
  pbuf[14].ui32[0] = 0x70000000;          // DMA end
}

/////////////////////////////////////////////////////////////////////////////

static int framecount;

static u_long128 *clutswap = NULL;

/////////////////////////////////////////////////////////////////////////////
//
// Set the color palette (768-byte VGA-style)
//
void gsvga_setpalette(unsigned char *pal) {
  // twiddle clut into our own temp buffer
  unsigned i, j;
  unsigned char *dest = (unsigned char*)clutswap;
#define COPYCOLOR {dest[0]=pal[0];dest[1]=pal[1];dest[2]=pal[2];dest[3]=0x80;}
  for(i = 0; i < 256; i += 32) {
    for(j = 0; j < 8; j++, pal+=3, dest+=4) { COPYCOLOR; }
    for(j = 0; j < 8; j++, pal+=3, dest+=4) { dest+=32;COPYCOLOR;dest-=32; }
    for(j = 0; j < 8; j++, pal+=3, dest+=4) { dest-=32;COPYCOLOR;dest+=32; }
    for(j = 0; j < 8; j++, pal+=3, dest+=4) { COPYCOLOR; }
  }
}

/////////////////////////////////////////////////////////////////////////////
//
// Emergency shutdown using DevCons
//
#define OFFX (((4096-320)/2)<<4)
#define OFFY (((4096-224)/2)<<4)
extern void sceDevConsLocate(int cd, u_int lx, u_int ly);

void gsvga_shutdown(const char *msg) {
  int console;

  sceGsSyncV(0);
  FlushCache(WRITEBACK_DCACHE);
  sceGsSwapDBuff(&db, 0);
  sceGsSyncPath(0, 0);

  sceDevConsInit();
  console = sceDevConsOpen(OFFX + (16<<4), OFFY + (24<<4), 32, 30);
  sceDevConsClear(console);
  sceDevConsLocate(console, 0, 0);
  sceDevConsPrintf(console, msg);
  sceDevConsDraw(console);
  sceGsSyncPath(0,0);

  sceGsSyncV(0);
  FlushCache(WRITEBACK_DCACHE);
  sceGsSwapDBuff(&db, 1);
  sceGsSyncPath(0, 0);

  for(;;);
}

/////////////////////////////////////////////////////////////////////////////
//
// Copy a frame to the screen (also vsyncs)
// frame must be on a 16-byte boundary
//
void gsvga_draw(unsigned char *frame) {
  sceGsLoadImage cp, tp;

  // load texture
  sceGsSetDefLoadImage(&tp, INDEXIMG_ADDRESS, 6, SCE_GS_PSMT8, 0, 0, 320, 224) ;
  FlushCache(WRITEBACK_DCACHE);
  sceGsExecLoadImage(&tp, (void*)frame);
  sceGsSyncPath(0, 0);

  // load clut
  sceGsSetDefLoadImage(&cp, CLUT_ADDRESS, 1, SCE_GS_PSMCT32, 0, 0, 16, 16); // PSMCT32
  FlushCache(WRITEBACK_DCACHE);
  sceGsExecLoadImage(&cp, clutswap);
  sceGsSyncPath(0, 0);

  // draw textured quad
  FlushCache(WRITEBACK_DCACHE);
  sceDmaSend(d2, packet_buf);
  sceGsSyncPath(0, 0);

  framecount++;

  // do file cache stuff while here
  filecache_process();

  // swap screen buffers
  sceGsSyncV(0);
  FlushCache(WRITEBACK_DCACHE);
  sceGsSwapDBuff(&db, (framecount) & 1);
  sceGsSyncPath(0, 0);

}

/////////////////////////////////////////////////////////////////////////////
//
// Initialization
//
void gsvga_init(void) {
  framecount = 0;

  sceDmaReset(1);
  sceGsResetPath();
  d2 = sceDmaGetChan(2);

  sceGsResetGraph(0, SCE_GS_NOINTERLACE, SCE_GS_NTSC, SCE_GS_FRAME);
  sceGsSetDefDBuff(&db, SCE_GS_PSMCT32, 320, 224, 0, 0, 0);

  init_packet(packet_buf);

  clutswap = tracemalloc("gsvga_init", 256*4);
  memset(clutswap, 0, 256*4);

  // clear the screen, prime the double-buffering
  { void *tmp = tracemalloc("gsvga_init", 320*240);
    if(tmp) {
      memset(tmp, 0, 320*240);
      gsvga_draw(tmp);
      tracefree(tmp);
    }
  }
}

/////////////////////////////////////////////////////////////////////////////

